# -*- coding: utf-8 -*-

# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance

"""This module contains classes and functionality relating to Sonos Groups."""

from __future__ import unicode_literals


class ZoneGroup(object):

    """
    A class representing a Sonos Group. It looks like this::

        ZoneGroup(
            uid='RINCON_000FD584236D01400:58',
            coordinator=SoCo("192.168.1.101"),
            members={SoCo("192.168.1.101"), SoCo("192.168.1.102")}
        )


    Any SoCo instance can tell you what group it is in::


        >>> device = soco.discovery.any_soco()
        >>> device.group
        ZoneGroup(
            uid='RINCON_000FD584236D01400:58',
            coordinator=SoCo("192.168.1.101"),
            members={SoCo("192.168.1.101"), SoCo("192.168.1.102")}
        )

    From there, you can find the coordinator for the current group::

        >>> device.group.coordinator
        SoCo("192.168.1.101")

    or, for example, its name::

        >>> device.group.coordinator.player_name
        Kitchen

    or a set of the members::

        >>> device.group.members
        {SoCo("192.168.1.101"), SoCo("192.168.1.102")}

    For convenience, ZoneGroup is also a container::

        >>> for player in device.group:
        ...   print player.player_name
        Living Room
        Kitchen

    If you need it, you can get an iterator over all groups on the network::

        >>> device.all_groups
        <generator object all_groups at 0x108cf0c30>

    A consistent readable label for the group members can be returned with
    the `label` and `short_label` properties.

    Properties are available to get and set the group `volume` and the group
    `mute` state, and the `set_relative_volume()` method can be used to make
    relative adjustments to the group volume, e.g.:

        >>> device.group.volume = 25
        >>> device.group.volume
        25
        >>> device.group.set_relative_volume(-10)
        15
        >>> device.group.mute
        >>> False
        >>> device.group.mute = True
        >>> device.group.mute
        True
    """

    def __init__(self, uid, coordinator, members=None):
        """
        Args:
            uid (str): The unique Sonos ID for this group, eg
                ``RINCON_000FD584236D01400:5``.
            coordinator (SoCo): The SoCo instance representing the coordinator
                of this group.
            members (Iterable[SoCo]): An iterable containing SoCo instances
                which represent the members of this group.
        """
        #: The unique Sonos ID for this group
        self.uid = uid
        #: The `SoCo` instance which coordinates this group
        self.coordinator = coordinator
        if members is not None:
            #: A set of `SoCo` instances which are members of the group
            self.members = set(members)
        else:
            self.members = set()

    def __iter__(self):
        return self.members.__iter__()

    def __contains__(self, member):
        return member in self.members

    def __repr__(self):
        return "{0}(uid='{1}', coordinator={2!r}, members={3!r})".format(
            self.__class__.__name__, self.uid, self.coordinator, self.members)

    @property
    def label(self):
        """str: A description of the group.

            >>> device.group.label
            'Kitchen, Living Room'
        """
        group_names = sorted([m.player_name for m in self.members])
        return ", ".join(group_names)

    @property
    def short_label(self):
        """str: A short description of the group.

        >>> device.group.short_label
        'Kitchen + 1'
        """
        group_names = sorted([m.player_name for m in self.members])
        group_label = group_names[0]
        if len(group_names) > 1:
            group_label += " + {}".format(len(group_names) - 1)
        return group_label

    @property
    def volume(self):
        """int: The volume of the group.

        An integer between 0 and 100.
        """
        response = self.coordinator.groupRenderingControl.GetGroupVolume([
            ('InstanceID', 0)
        ])
        return int(response['CurrentVolume'])

    @volume.setter
    def volume(self, group_volume):
        group_volume = int(group_volume)
        group_volume = max(0, min(group_volume, 100))  # Coerce in range
        self.coordinator.groupRenderingControl.SetGroupVolume([
            ('InstanceID', 0),
            ('DesiredVolume', group_volume)
        ])

    @property
    def mute(self):
        """bool: The mute state for the group.

        True or False.
        """
        response = self.coordinator.groupRenderingControl.GetGroupMute([
            ('InstanceID', 0)
        ])
        mute_state = response['CurrentMute']
        return bool(int(mute_state))

    @mute.setter
    def mute(self, group_mute):
        mute_value = '1' if group_mute else '0'
        self.coordinator.groupRenderingControl.SetGroupMute([
            ('InstanceID', 0),
            ('DesiredMute', mute_value)
        ])

    def set_relative_volume(self, relative_group_volume):
        """Adjust the group volume up or down by a relative amount.

        If the adjustment causes the volume to overshoot the maximum value
        of 100, the volume will be set to 100. If the adjustment causes the
        volume to undershoot the minimum value of 0, the volume will be set
        to 0.

        Note that this method is an alternative to using addition and
        subtraction assignment operators (+=, -=) on the `volume` property
        of a `ZoneGroup` instance. These operators perform the same function
        as `set_relative_volume()` but require two network calls per
        operation instead of one.

        Args:
            relative_group_volume (int): The relative volume adjustment. Can be
                positive or negative.

        Returns:
            int: The new group volume setting.

        Raises:
            ValueError: If `relative_group_volume` cannot be cast as
                an integer.
        """
        relative_group_volume = int(relative_group_volume)
        # Sonos automatically handles out-of-range values.
        resp = self.coordinator.groupRenderingControl.SetRelativeGroupVolume([
            ('InstanceID', 0),
            ('Adjustment', relative_group_volume)
        ])
        return int(resp['NewVolume'])



# -*- coding: utf-8 -*-
# pylint: disable=R0201,E0711

# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance

"""This is the __init__ module for the plugins.

It contains the base class for all plugins
"""

import logging
import importlib


_LOG = logging.getLogger(__name__)


class SoCoPlugin(object):

    """The base class for SoCo plugins."""

    def __init__(self, soco):
        cls = self.__class__.__name__
        _LOG.debug('Initializing SoCo plugin %s', cls)
        self.soco = soco

    @property
    def name(self):
        """ human-readable name of the plugin """
        raise NotImplementedError('Plugins should overwrite the name property')

    @classmethod
    def from_name(cls, fullname, soco, *args, **kwargs):
        """Instantiate a plugin by its full name."""

        _LOG.debug('Loading plugin %s', fullname)

        parts = fullname.split('.')
        modname = '.'.join(parts[:-1])
        clsname = parts[-1]

        mod = importlib.import_module(modname)
        class_ = getattr(mod, clsname)

        _LOG.debug('Loaded class %s', class_)

        return class_(soco, *args, **kwargs)



"""py.test hooks.

Add the --ip command line option, and skip all tests marked the with
'integration' marker unless the option is included
"""
import pytest


def pytest_addoption(parser):
    """ Add the --ip commandline option """
    parser.addoption(
        '--ip',
        type=str,
        default=None,
        action='store',
        dest='IP',
        help='the IP address for the zone to be used for the integration tests'
    )


def pytest_runtest_setup(item):
    """Skip tests marked 'integration' unless an ip address is given."""
    if "integration" in item.keywords and not item.config.getoption("--ip"):
        pytest.skip("use --ip and an ip address to run integration tests.")


class Helpers(object):
    """Test helper functions"""

    @staticmethod
    def compare_xml(element1, element2):
        # print()
        # print(element1.tag, "TAG", element2.tag)
        # print(element1.attrib, element2.attrib)
        if element1.tag != element2.tag:
            return False

        # print(element1.attrib, element2.attrib)
        if element1.attrib != element2.attrib:
            return False

        # print(len(element1), len(element2))
        if len(element1) != len(element2):
            return False

        for subelement1, subelement2 in zip(element1, element2):
            if not Helpers.compare_xml(subelement1, subelement2):
                return False

        return True


@pytest.fixture
def helpers():
    return Helpers



# -*- coding: utf-8 -*-
# pylint: disable=fixme

# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance

"""Classes for handling SoCo's basic SOAP requirements.

This module does not handle anything like the full `SOAP Specification
<http://www.w3.org/TR/soap/>`_ , but is enough for SoCo's needs. Sonos uses
SOAP for UPnP communications, and for communication with third party music
services.
"""

# The state of Python's SOAP libraries is poor. In any event, the two main
# libraries, PySimpleSOAP and SUDS (or the more up-to-date SUDS-Jurko),
# are too complex for our needs. SUDS requires a WSDL file to be parsed,
# and although SONOS provides one in relation to music services (at
# http://musicpartners.sonos.com/sites/default/files/Sonos.wsdl) the various
# music services themselves provide buggy, incomplete or old
# implementations which cause SUDS to break. PySimpleSOAP can work without a
# WSDL file, but contains various bugs which mean that we would have to use
# a patched version (upstream releases are infrequent).  Since SONOS only
# appears to use basic SOAP features, and after experimenting with other
# libraries, it seems best to write our own.

# Some is the same as that in services.py.
# TODO: refactor services.py to depend on this code

from __future__ import (
    absolute_import, unicode_literals
)

import logging
from xml.sax.saxutils import escape

import requests

from .exceptions import SoCoException
from .utils import prettify
from .xml import XML

_LOG = logging.getLogger(__name__)


class SoapFault(SoCoException):

    """An exception encapsulating a SOAP Fault."""

    def __init__(self, faultcode, faultstring, detail=None):
        """
        Args:
            faultcode (str): The SOAP faultcode.
            faultstring (str): The SOAP faultstring.
            detail (:obj:`~xml.etree.ElementTree.Element`): The SOAP fault
                detail, as an ElementTree
                :obj:`~xml.etree.ElementTree.Element`. Defaults to `None`.
        """
        self.faultcode = faultcode
        self.faultstring = faultstring
        self.detail = detail
        self.detail_string = XML.tostring(detail) if detail is not None else ''
        super(SoapFault, self).__init__(faultcode, faultstring)

    def __str__(self):
        return '%s: %s' % (self.faultcode, self.faultstring)

    def __repr__(self):
        return "SoapFault(faultcode={}, faultstring={}, detail={})".format(
            repr(self.faultcode),
            repr(self.faultstring),
            repr(self.detail)
        )


# Sonos uses SOAP to send commands in the RPC form. A complete RPC SOAP
# message should look something like this. See generally
# http://www.w3.org/TR/2000/NOTE-SOAP-20000508/

# POST Endpoint URL HTTP/1.1
# HOST: Host of Endpoint URL:port
# CONTENT-LENGTH: bytes in body
# CONTENT-TYPE: text/xml; charset="utf-8"
# SOAPACTION: URI
#
# <?xml version="1.0"?>
# <s:Envelope
#   xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"
#   s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
#   <s:Header>
#       </Header elements go here>
#   </s:Header>
#   <s:Body>
#       <ns:MethodName xmlns:ns="MethodNamespace>"
#           <param1>value</param1>
#           ...
#           <param_n>value</param_n>
#       </ns:MethodName>
#   </s:Body>
# </s:Envelope>

# pylint: disable=too-many-instance-attributes, too-many-arguments


class SoapMessage(object):

    """A SOAP Message representing a remote procedure call.

    Uses the `Requests <http://www.python-requests.org/en/latest/>`_ library
    for communication with a SOAP server.
    """

    def __init__(self, endpoint, method, parameters=None, http_headers=None,
                 soap_action=None, soap_header=None, namespace=None,
                 **request_args):
        """
        Args:
            endpoint (str): The SOAP endpoint URL for this client.
            method (str): The name of the method to call.
            parameters (list): A list of (name, value) tuples containing
                the parameters to pass to the method. Default `None`.
            http_headers (dict): A dict in the form {'Header': 'Value,..}
                containing http headers to use for the http request.
                Content-type and SOAPACTION headers will be created
                automatically, so do not include them here. Use this, for
                example, to set a user-agent.
            soap_action (str): The value of the SOAPACTION header.
                Default 'None`.
            soap_header (str): A string representation of the XML to be
                used for the SOAP Header. Default `None`.
            namespace (str): The namespace URI to use for the method and
                parameters. `None`, by default.
            **request_args: Other keyword parameters will be passed to the
                Requests request which is used to handle the http
                communication. For example, a timeout value can be set.
        """
        self.endpoint = endpoint
        self.method = method
        self.parameters = [] if parameters is None else parameters
        self.http_headers = http_headers
        self.soap_action = soap_action
        self.soap_header = soap_header
        self.namespace = namespace
        self.request_args = request_args

    # pylint:disable=no-self-use
    def prepare_headers(self, http_headers, soap_action):
        """Prepare the http headers for sending.

        Add the SOAPACTION header to the others.

        Args:
            http_headers (dict): A dict in the form {'Header': 'Value,..}
                containing http headers to use for the http request.
            soap_action (str): The value of the SOAPACTION header.

        Returns:
            dict: headers including the SOAPACTION header.
        """

        headers = {'Content-Type': 'text/xml; charset="utf-8"'}
        if soap_action is not None:
            headers.update({'SOAPACTION': '"{}"'.format(soap_action)})
        if http_headers is not None:
            headers.update(http_headers)
        return headers

    def prepare_soap_header(self, soap_header):
        """Prepare the SOAP header for sending.

        Wraps the soap header in appropriate tags.

        Args:
            soap_header (str): A string representation of the XML to be
                used for the SOAP Header

        Returns:
            str: The soap header wrapped in appropriate tags.
        """

        if soap_header is not None:
            return '<s:Header>{}</s:Header>'.format(soap_header)
        else:
            return ''

    def prepare_soap_body(self, method, parameters, namespace):
        """Prepare the SOAP message body for sending.

        Args:
            method (str): The name of the method to call.
            parameters (list): A list of (name, value) tuples containing
                the parameters to pass to the method.
            namespace (str): tThe XML namespace to use for the method.

        Returns:
            str: A properly formatted SOAP Body.
        """

        tags = []
        for name, value in parameters:
            tag = "<{name}>{value}</{name}>".format(
                name=name, value=escape("%s" % value, {'"': "&quot;"}))
            # % converts to unicode because we are using unicode literals.
            # Avoids use of 'unicode' function which does not exist in python 3
            tags.append(tag)

        wrapped_params = "".join(tags)
        # Prepare the SOAP Body
        if namespace is not None:
            soap_body = (
                '<{method} xmlns="{namespace}">'
                '{params}'
                '</{method}>'.format(
                    method=method, params=wrapped_params,
                    namespace=namespace
                ))
        else:
            soap_body = (
                '<{method}>'
                '{params}'
                '</{method}>'.format(
                    method=method, params=wrapped_params
                ))

        return soap_body

    def prepare_soap_envelope(self, prepared_soap_header, prepared_soap_body):
        """Prepare the SOAP Envelope for sending.

        Args:
            prepared_soap_header (str): A SOAP Header prepared by
                `prepare_soap_header`
            prepared_soap_body (str): A SOAP Body prepared by
                `prepare_soap_body`

        Returns:
            str: A prepared SOAP Envelope
        """

        # pylint: disable=bad-continuation
        soap_env_template = (
            '<?xml version="1.0"?>'
            '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"'
            ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">'
            '{soap_header}'
            '<s:Body>'
            '{soap_body}'
            '</s:Body>'
            '</s:Envelope>')  # noqa PEP8
        return soap_env_template.format(
            soap_header=prepared_soap_header,
            soap_body=prepared_soap_body)

    def prepare(self):
        """Prepare the SOAP message for sending to the server."""
        headers = self.prepare_headers(self.http_headers, self.soap_action)

        soap_header = self.prepare_soap_header(self.soap_header)
        soap_body = self.prepare_soap_body(
            self.method, self.parameters, self.namespace
        )
        data = self.prepare_soap_envelope(soap_header, soap_body)
        return (headers, data)

    def call(self):
        """Call the SOAP method on the server.

        Returns:
            str: the decapusulated SOAP response from the server,
                still encoded as utf-8.

        Raises:
             SoapFault: if a SOAP error occurs.
             ~requests.exceptions.HTTPError: if an http error occurs.

        """

        headers, data = self.prepare()

        # Check log level before logging XML, since prettifying it is
        # expensive
        if _LOG.isEnabledFor(logging.DEBUG):
            _LOG.debug("Sending %s, %s", headers, prettify(data))

        response = requests.post(
            self.endpoint,
            headers=headers,
            data=data.encode('utf-8'),
            **self.request_args
        )
        _LOG.debug("Received %s, %s", response.headers, response.text)
        status = response.status_code
        if status == 200:
            # The response is good. Extract the Body
            tree = XML.fromstring(response.content)
            # Get the first child of the <Body> tag. NB There should only be
            # one if the RPC standard is followed.
            body = tree.find(
                "{http://schemas.xmlsoap.org/soap/envelope/}Body")[0]
            return body
        elif status == 500:
            # We probably have a SOAP Fault
            tree = XML.fromstring(response.content)
            fault = tree.find(
                './/{http://schemas.xmlsoap.org/soap/envelope/}Fault'
            )
            if fault is None:
                # Not a SOAP fault. Must be something else.
                response.raise_for_status()
            faultcode = fault.findtext("faultcode")
            faultstring = fault.findtext("faultstring")
            faultdetail = fault.find("detail")
            raise SoapFault(faultcode, faultstring, faultdetail)
        else:
            # Something else has gone wrong. Probably a network error. Let
            # Requests handle it
            response.raise_for_status()

        return None



# -*- coding: utf-8 -*-

"""A SoCo fork with fixes for Home Assistant."""

# There is no need for all strings here to be unicode, and Py2 cannot import
# modules with unicode names so do not use from __future__ import
# unicode_literals
# https://github.com/SoCo/SoCo/issues/98
#


import logging
import warnings
import sys

from .core import SoCo
from .discovery import discover, discover_thread
from .exceptions import SoCoException, UnknownSoCoException

# Will be parsed by setup.py to determine package metadata
__author__ = 'Anders Melchiorsen <amelchio@nogoto.net>'
# Please add the suffix "+" to the version after release, to make it
# possible infer whether in development code from the version string
__version__ = '0.0.35'
__website__ = 'https://github.com/amelchio/pysonos'
__license__ = 'MIT License'

# You really should not `import *` - it is poor practice
# but if you do, here is what you get:
__all__ = [
    'discover',
    'discover_thread',
    'SoCo',
    'SoCoException',
    'UnknownSoCoException',
]

# http://docs.python.org/2/howto/logging.html#library-config
# Avoids spurious error messages if no logger is configured by the user

logging.getLogger(__name__).addHandler(logging.NullHandler())

if sys.version_info.major < 3:
    warnings.warn(
        "Version 0.19 of SoCo is the last to support Python 2.7",
        stacklevel=2,
    )



# -*- coding: utf-8 -*-
"""Tests for the utils module."""

from __future__ import unicode_literals

from pysonos.utils import deprecated


# Deprecation decorator


def test_deprecation(recwarn):

    @deprecated('0.7')
    def dummy(args):
        """My docs."""
        pass

    @deprecated('0.8', 'better_function', '0.12')
    def dummy2(args):
        """My docs."""
        pass

    assert dummy.__doc__ == "My docs.\n\n  .. deprecated:: 0.7\n"
    assert dummy2.__doc__ == "My docs.\n\n  .. deprecated:: 0.8\n\n"\
                             "     Will be removed in version 0.12.\n" \
                             "     Use `better_function` instead."
    dummy(3)
    w = recwarn.pop()
    assert str(w.message) == 'Call to deprecated function dummy.'
    dummy2(4)
    w = recwarn.pop()
    assert str(w.message) == "Call to deprecated function dummy2. Will be " \
                             "removed in version 0.12. Use " \
                             "better_function instead."
    assert w.filename
    assert w.lineno



# -*- coding: utf-8 -*-

"""This module contains configuration variables.

They may be set by your code as follows::

    from soco import config
    ...
    config.VARIABLE = value
"""

from __future__ import unicode_literals


SOCO_CLASS = None
"""The class object to use when `SoCo` instances are created.

Specify the actual callable class object here, not a string. If `None`,
the default SoCo class will be used. Must be set before any instances are
created, or it will have unpredictable effects.
"""


CACHE_ENABLED = True
"""Is the cache enabled?

If `True` (the default), some caching of network requests will take place.

See also:
    The :mod:`soco.cache` module.
"""


EVENT_ADVERTISE_IP = None
"""The IP on which to advertise to Sonos.

The default of None means that the relevant IP address will be detected
automatically.

See also:
    The :mod:`soco.events_base` module.
"""

EVENT_LISTENER_IP = None
"""The IP on which the event listener listens.

The default of None means that the relevant IP address will be detected
automatically.

See also:
    The :mod:`soco.events_base` module.
"""


EVENT_LISTENER_PORT = 1400
"""The port on which the event listener listens.

The default is 1400. You must set this before subscribing to any events.

See also:
    The :mod:`soco.events_base` module.
"""

EVENTS_MODULE = None
"""The events module to be used by the :mod:`soco.services` module.

The default of None means the :mod:`soco.events` module will be used.

See also:
    The :mod:`soco.events` and :mod:`soco.events_twisted` modules.
"""



# -*- coding: utf-8 -*-
# pylint: disable=too-many-instance-attributes

# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance

"""Functionality to support saving and restoring the current Sonos state.

This is useful for scenarios such as when you want to switch to radio
or an announcement and then back again to what was playing previously.

Warning:
    Sonos has introduced control via Amazon Alexa. A new cloud queue is
    created and at present there appears no way to restart this
    queue from snapshot. Currently if a cloud queue was playing it will
    not restart.

Warning:
    This class is designed to be created used and destroyed. It is not
    designed to be reused or long lived. The init sets up defaults for
    one use.
"""


class Snapshot(object):
    """A snapshot of the current state.

    Note:
        This does not change anything to do with the configuration
        such as which group the speaker is in, just settings that impact
        what is playing, or how it is played.

        List of sources that may be playing using root of media_uri:

        | ``x-rincon-queue``: playing from Queue
        | ``x-sonosapi-stream``: playing a stream (eg radio)
        | ``x-file-cifs``: playing file
        | ``x-rincon``: slave zone (only change volume etc. rest from
          coordinator)
    """

    def __init__(self, device, snapshot_queue=False):
        """
        Args:
            device (SoCo): The device to snapshot
            snapshot_queue (bool): Whether the queue should be snapshotted.
                Defaults to `False`.

        Warning:
            It is strongly advised that you do not snapshot the queue unless
            you really need to as it takes a very long time to restore large
            queues as it is done one track at a time.
        """
        # The device that will be snapshotted
        self.device = device

        # The values that will be stored
        # For all zones:
        self.media_uri = None
        self.is_coordinator = False
        self.is_playing_queue = False
        self.is_playing_cloud_queue = False

        self.volume = None
        self.mute = None
        self.bass = None
        self.treble = None
        self.loudness = None

        # For coordinator zone playing from Queue:
        self.play_mode = None
        self.cross_fade = None
        self.playlist_position = 0
        self.track_position = None

        # For coordinator zone playing a Stream:
        self.media_metadata = None

        # For all coordinator zones
        self.transport_state = None

        self.queue = None
        # Only set the queue as a list if we are going to save it
        if snapshot_queue:
            self.queue = []

    def snapshot(self):
        """Record and store the current state of a device.

        Returns:
            bool: `True` if the device is a coordinator, `False` otherwise.
                Useful for determining whether playing an alert on a device
                will ungroup it.
        """
        # get if device coordinator (or slave) True (or False)
        self.is_coordinator = self.device.is_coordinator

        # Get information about the currently playing media
        media_info = self.device.avTransport.GetMediaInfo([('InstanceID', 0)])
        self.media_uri = media_info['CurrentURI']
        # Extract source from media uri - below some media URI value examples:
        #  'x-rincon-queue:RINCON_000E5859E49601400#0'
        #       - playing a local queue always #0 for local queue)
        #
        #  'x-rincon-queue:RINCON_000E5859E49601400#6'
        #       - playing a cloud queue where #x changes with each queue)
        #
        #  -'x-rincon:RINCON_000E5859E49601400'
        #       - a slave player pointing to coordinator player

        if self.media_uri.split(':')[0] == 'x-rincon-queue':
            # The pylint error below is a false positive, see about removing it
            # in the future
            # pylint: disable=simplifiable-if-statement
            if self.media_uri.split('#')[1] == '0':
                # playing local queue
                self.is_playing_queue = True
            else:
                # playing cloud queue - started from Alexa
                self.is_playing_cloud_queue = True

        # Save the volume, mute and other sound settings
        self.volume = self.device.volume
        self.mute = self.device.mute
        self.bass = self.device.bass
        self.treble = self.device.treble
        self.loudness = self.device.loudness

        # get details required for what's playing:
        if self.is_playing_queue:
            # playing from queue - save repeat, random, cross fade, track, etc.
            self.play_mode = self.device.play_mode
            self.cross_fade = self.device.cross_fade

            # Get information about the currently playing track
            track_info = self.device.get_current_track_info()
            if track_info is not None:
                position = track_info['playlist_position']
                if position != "":
                    # save as integer
                    self.playlist_position = int(position)
                self.track_position = track_info['position']
        else:
            # playing from a stream - save media metadata
            self.media_metadata = media_info['CurrentURIMetaData']

        # Work out what the playing state is - if a coordinator
        if self.is_coordinator:
            transport_info = self.device.get_current_transport_info()
            if transport_info is not None:
                self.transport_state = transport_info[
                    'current_transport_state']

        # Save of the current queue if we need to
        self._save_queue()

        # return if device is a coordinator (helps usage)
        return self.is_coordinator

    # pylint: disable=too-many-branches
    def restore(self, fade=False):
        """Restore the state of a device to that which was previously saved.

        For coordinator devices restore everything. For slave devices
        only restore volume etc., not transport info (transport info
        comes from the slave's coordinator).

        Args:
            fade (bool): Whether volume should be faded up on restore.
        """

        try:
            if self.is_coordinator:
                self._restore_coordinator()
        finally:
            self._restore_volume(fade)

        # Now everything is set, see if we need to be playing, stopped
        # or paused ( only for coordinators)
        if self.is_coordinator:
            if self.transport_state == 'PLAYING':
                self.device.play()
            elif self.transport_state == 'STOPPED':
                self.device.stop()

    def _restore_coordinator(self):
        """Do the coordinator-only part of the restore."""
        # Start by ensuring that the speaker is paused as we don't want
        # things all rolling back when we are changing them, as this could
        # include things like audio
        transport_info = self.device.get_current_transport_info()
        if transport_info is not None:
            if transport_info['current_transport_state'] == 'PLAYING':
                self.device.pause()

        # Check if the queue should be restored
        self._restore_queue()

        # Reinstate what was playing

        if self.is_playing_queue and self.playlist_position > 0:
            # was playing from playlist

            if self.playlist_position is not None:
                # The position in the playlist returned by
                # get_current_track_info starts at 1, but when
                # playing from playlist, the index starts at 0
                # if position > 0:
                self.playlist_position -= 1
                self.device.play_from_queue(self.playlist_position, False)

            if self.track_position is not None:
                if self.track_position != "":
                    self.device.seek(self.track_position)

            # reinstate track, position, play mode, cross fade
            # Need to make sure there is a proper track selected first
            self.device.play_mode = self.play_mode
            self.device.cross_fade = self.cross_fade

        elif self.is_playing_cloud_queue:
            # was playing a cloud queue started by Alexa
            # No way yet to re-start this so prevent it throwing an error!
            pass

        else:
            # was playing a stream (radio station, file, or nothing)
            # reinstate uri and meta data
            if self.media_uri != "":
                self.device.play_uri(
                    self.media_uri, self.media_metadata, start=False)

    def _restore_volume(self, fade):
        """Reinstate volume.

        Args:
            fade (bool): Whether volume should be faded up on restore.
        """
        self.device.mute = self.mute

        # Can only change volume on device with fixed volume set to False
        # otherwise get uPnP error, so check first. Before issuing a network
        # command to check, fixed volume always has volume set to 100.
        # So only checked fixed volume if volume is 100.
        if self.volume == 100:
            fixed_vol = self.device.renderingControl.GetOutputFixed(
                [('InstanceID', 0)])['CurrentFixed']
        else:
            fixed_vol = False

        # now set volume if not fixed
        if not fixed_vol:
            self.device.bass = self.bass
            self.device.treble = self.treble
            self.device.loudness = self.loudness

            if fade:
                # if fade requested in restore
                # set volume to 0 then fade up to saved volume (non blocking)
                self.device.volume = 0
                self.device.ramp_to_volume(self.volume)
            else:
                # set volume
                self.device.volume = self.volume

    def _save_queue(self):
        """Save the current state of the queue."""
        if self.queue is not None:
            # Maximum batch is 486, anything larger will still only
            # return 486
            batch_size = 400
            total = 0
            num_return = batch_size

            # Need to get all the tracks in batches, but Only get the next
            # batch if all the items requested were in the last batch
            while num_return == batch_size:
                queue_items = self.device.get_queue(total, batch_size)
                # Check how many entries were returned
                num_return = len(queue_items)
                # Make sure the queue is not empty
                if num_return > 0:
                    self.queue.append(queue_items)
                # Update the total that have been processed
                total = total + num_return

    def _restore_queue(self):
        """Restore the previous state of the queue.

        Note:
            The restore currently adds the items back into the queue
            using the URI, for items the Sonos system already knows about
            this is OK, but for other items, they may be missing some of
            their metadata as it will not be automatically picked up.
        """
        if self.queue is not None:
            # Clear the queue so that it can be reset
            self.device.clear_queue()
            # Now loop around all the queue entries adding them
            for queue_group in self.queue:
                for queue_item in queue_group:
                    self.device.add_uri_to_queue(queue_item.uri)

    def __enter__(self):
        self.snapshot()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.restore()



#!/usr/bin/env python
"""
Play a selected favorite radio station from the TuneIn service on your Sonos
system, or a short clip from each favorite.

Pass an IP address as the first command-line argument. If you specify a preset
as the second argument, that radio station will be played. Otherwise a short
clip from each preset will be played.

"""

from __future__ import unicode_literals
from __future__ import print_function

import sys
import time

from soco import SoCo

meta_template = """
<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/"
    xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
    xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
    xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
    <item id="R:0/0/0" parentID="R:0/0" restricted="true">
        <dc:title>{title}</dc:title>
        <upnp:class>object.item.audioItem.audioBroadcast</upnp:class>
        <desc id="cdudn" nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
            {service}
        </desc>
    </item>
</DIDL-Lite>' """

tunein_service = "SA_RINCON65031_"

if __name__ == "__main__":

    if len(sys.argv) < 2:
        print("Please pass the IP address of a Zone Player as the first argument")
        sys.exit()

    speaker_ip = sys.argv[1]
    preset = 0
    limit = 12

    if len(sys.argv) == 3:
        preset = int(sys.argv[2]) - 1
        limit = 1

    mySonos = SoCo(speaker_ip)

    if mySonos:
        stations = mySonos.get_favorite_radio_stations(preset, limit)
        print(
            "returned %s of a possible %s radio stations:"
            % (stations["returned"], stations["total"])
        )
    for station in stations["favorites"]:
        print(station["title"])
        uri = station["uri"]
        # TODO seems at least & needs to be escaped - should move this to
        # play_uri and maybe escape other chars.
        uri = uri.replace("&", "&amp;")

        metadata = meta_template.format(title=station["title"], service=tunein_service)

        print(mySonos.play_uri(uri, metadata))

        if len(sys.argv) == 2:
            time.sleep(10)



# -*- coding: utf-8 -*-

"""This module implements 'quirks' for the DIDL-Lite data structures

A quirk, in this context, means that a specific music service does not follow
a specific part of the DIDL-Lite specification. In order not to clutter the
primary implementation of DIDL-Lite for SoCo (in :mod:`soco.data_structures`)
up with all these service specific exception, they are implemented separately
in this module. Besides from keeping the main implementation clean and
following the specification, this has the added advantage of making it easier
to track how many quiks are out there.

The implementation of the quirks at this point is just a single function which
applies quirks to the DIDL-Lite resources, with the options of adding one that
applies them to DIDL-Lite objects.

"""

import logging


_LOG = logging.getLogger(__name__)


def apply_resource_quirks(resource):
    """Apply DIDL-Lite resource quirks"""
    # At least two music service (Spotify Direct and Amazon in conjunction
    # with Alexa) has been known not to supply the mandatory protocolInfo, so
    # if it is missing supply a dummy one
    if 'protocolInfo' not in resource.attrib:
        protocol_info = 'DUMMY_ADDED_BY_QUIRK'
        # For Spotify direct we have a better idea what it should be, since it
        # is included in the main element text
        if resource.text and resource.text.startswith("x-sonos-spotify"):
            protocol_info = "sonos.com-spotify:*:audio/x-spotify.*"

        _LOG.debug(
            "Resource quirk applied for missing protocolInfo, setting to '%s'",
            protocol_info,
        )
        resource.set('protocolInfo', protocol_info)

        if not resource.text:
            resource.text = ""
    return resource



#!/usr/bin/env python

import io
import re
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    # Code from here: https://pytest.org/latest/goodpractises.html

    def finalize_options(self):
        TestCommand.finalize_options(self)
        # we don't run integration tests which need an actual Sonos device
        self.test_args = ['-m', 'not integration']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


src = io.open('pysonos/__init__.py', encoding='utf-8').read()
metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", src))
docstrings = re.findall('"""(.*?)"""', src, re.MULTILINE | re.DOTALL)

NAME = 'pysonos'

PACKAGES = (
    'pysonos',
    'pysonos.plugins',
    'pysonos.music_services',
)

TEST_REQUIREMENTS = list(open('requirements-dev.txt'))
AUTHOR_EMAIL = metadata['author']
VERSION = metadata['version']
WEBSITE = metadata['website']
LICENSE = metadata['license']
DESCRIPTION = docstrings[0]

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Home Automation',
    'Topic :: Multimedia :: Sound/Audio',
    'Topic :: Multimedia :: Sound/Audio :: Players',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

with io.open('README.rst', encoding='utf-8') as file:
    LONG_DESCRIPTION = file.read()

# Extract name and e-mail ("Firstname Lastname <mail@example.org>")
AUTHOR, EMAIL = re.match(r'(.*) <(.*)>', AUTHOR_EMAIL).groups()

REQUIREMENTS = list(open('requirements.txt'))

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    license=LICENSE,
    url=WEBSITE,
    packages=PACKAGES,
    install_requires=REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    long_description=LONG_DESCRIPTION,
    cmdclass={'test': PyTest},
    classifiers=CLASSIFIERS,
)
